import sys

from typing import List, Optional
from argparse import ArgumentParser, Namespace, BooleanOptionalAction

from onep import commands

from .util import init_secret_storage, check_session, fatal
from .colors import *


def main(args: Optional[List[str]] = None) -> None:
    init_secret_storage()

    cli = parse_args(args)

    if cli.command is None:
        fatal("no command provided")

    if cli.command != "signin":
        session = check_session(cli.account)

        if session is None:
            sys.exit(1)

    assert session is not None

    if cli.command == "signin":
        commands.signin.signin(cli.account)
    elif cli.command == "vaults":
        commands.vaults.vaults(session, cli.json)
    elif cli.command == "vault":
        commands.vaults.vault(session, cli.json, cli.id)
    elif cli.command == "search":
        commands.search.search(session, cli.json, cli.vault, cli.tags, cli.term)
    elif cli.command == "show":
        commands.show.show(session, cli.json, cli.id, cli.tags, cli.fields, cli.otp, cli.select)
    elif cli.command == "create":
        commands.edit.create(session, cli.vault, cli.category, cli.title, cli.fields, cli.tags, cli.password_length, cli.symbols)
    elif cli.command == "edit":
        commands.edit.edit(session, cli.id, cli.fields, cli.tags, cli.delete, cli.password_length, cli.symbols)
    elif cli.command == "delete":
        commands.delete.delete(session, cli.archive, cli.id)
    elif cli.command == "download":
        commands.document.download(session, cli.title)
    elif cli.command == "upload":
        commands.document.upload(session, cli.vault, cli.title, cli.file, cli.filename)
    elif cli.command == "share":
        commands.share.share(session, cli.id, cli.time, cli.once)


def parse_args(args: Optional[List[str]]) -> Namespace:
    cli = ArgumentParser(prog="1p")
    cli.add_argument("-j", "--json", action="store_true", help="format output as JSON")
    cli.add_argument("account", metavar="ACCOUNT", help="shorthand of the 1Password account")

    commands = cli.add_subparsers(dest="command", metavar="COMMAND")

    commands.add_parser("signin", help="authenticate into a 1Password account")
    commands.add_parser("vaults", help="list available vaults")

    cmd_vault = commands.add_parser("vault", help="show information about a vault")
    cmd_vault.add_argument("id", metavar="ID")

    cmd_search = commands.add_parser("search", help="search entries matching provided term")
    cmd_search.add_argument("-v", "--vault", type=str, metavar="VAULT")
    cmd_search.add_argument("-t", "--tags", type=str, metavar="TAGS")
    cmd_search.add_argument("term", metavar="TERM", nargs="*", default=[])

    cmd_show = commands.add_parser("show", help="display an entry")
    cmd_show_target = cmd_show.add_mutually_exclusive_group(required=True)
    cmd_show_target.add_argument("id", metavar="ID", nargs="*", default=[])
    cmd_show_target.add_argument("-t", "--tags", type=str, metavar="TAGS")
    cmd_show.add_argument("-f", "--fields", type=str, metavar="FIELDS")
    cmd_show.add_argument("-o", "--otp", action="store_true")
    cmd_show.add_argument("-s", "--select", action="store_true")

    cmd_create = commands.add_parser("create", help="create an entry")
    cmd_create.add_argument("-v", "--vault", type=str, metavar="VAULT", required=True)
    cmd_create.add_argument("-c", "--category", type=str, metavar="CATEGORY", default="login")
    cmd_create.add_argument("-u", "--url", type=str, metavar="URL")
    cmd_create.add_argument("-t", "--tags", type=str, metavar="TAGS")
    cmd_create.add_argument("--password-length", type=int, default=32)
    cmd_create.add_argument("--symbols", type=bool, action=BooleanOptionalAction, default=True)
    cmd_create.add_argument("title", metavar="TITLE")
    cmd_create.add_argument("fields", metavar="FIELD", nargs="+")

    cmd_edit = commands.add_parser("edit", help="edit an entry")
    cmd_edit.add_argument("-t", "--tags", type=str, metavar="TAGS")
    cmd_edit.add_argument("-d", "--delete", type=str, action="append", metavar="FIELD")
    cmd_edit.add_argument("--password-length", type=int, default=32)
    cmd_edit.add_argument("--symbols", type=bool, action=BooleanOptionalAction, default=True)
    cmd_edit.add_argument("id", metavar="ID")
    cmd_edit.add_argument("fields", metavar="FIELD", nargs="*")

    cmd_delete = commands.add_parser("delete", help="delete an entry")
    cmd_delete.add_argument("-a", "--archive", action="store_true")
    cmd_delete.add_argument("id", metavar="ID")

    cmd_download = commands.add_parser("download", help="download a document")
    cmd_download.add_argument("title", metavar="TITLE")

    cmd_upload = commands.add_parser("upload", help="upload a document")
    cmd_upload.add_argument("-v", "--vault", type=str, metavar="VAULT", required=True)
    cmd_upload.add_argument("-f", "--filename", type=str, metavar="FILENAME")
    cmd_upload.add_argument("title", metavar="TITLE")
    cmd_upload.add_argument("file", metavar="FILE")

    cmd_share = commands.add_parser("share", help="get a shareable link to an item")
    cmd_share.add_argument("id", metavar="ID")
    cmd_share.add_argument("-t", "--time", type=str, metavar="EXPIRY")
    cmd_share.add_argument("-o", "--once", action="store_true")

    return cli.parse_args(args)



import json

from typing import List

from ..util import fatal, run, create_table


def search(session: str, to_json: bool, vault: str, tags: str, term: List[str]) -> None:
    args = ["item", "list", "--format=json"]

    if vault is not None:
        args.append(f"--vault={vault}")

    if tags is not None:
        args.append(f"--tags={tags}")

    (status, stdout, stderr) = run(args, session=session, json=to_json, silent=True)

    if not status:
        fatal(stderr)

    entries = json.loads(stdout)

    if term is not None:
        id = " ".join(term)
        entries = list(filter(lambda entry: id.lower() in entry["title"].lower(), entries))

    entries = list(map(lambda entry: [entry["id"], entry["title"]], sorted(entries, key=lambda entry: entry["title"])))  # type: ignore

    table = create_table(["ID", "TITLE"])
    table.add_rows(entries)

    print(table)



import sys

from typing import Optional

from ..util import fatal, run


def download(session: str, title: str) -> None:
    (status, stdout, stderr) = run(["document", "get", title], session=session, silent=True)

    if not status:
        fatal(stderr)

    print(stdout)


def upload(session: str, vault: str, title: str, file: str, filename: Optional[str]) -> None:
    args = ["document", "create", f"--vault={vault}", f"--title={title}", file]

    if filename is not None:
        args.append(f"--file-name={filename}")

    stdin = None

    if file == "-":
        stdin = sys.stdin.read()

    (status, _, stderr) = run(args, session=session, silent=True, stdin=stdin)

    if not status:
        fatal(stderr)



import json

from ..util import fatal, run


def delete(session: str, archive: bool, id: str) -> None:
    (status, _, stderr) = run(["item", "get", "--format=json", id], session=session, silent=True)

    if not status:
        fatal(stderr)

    args = ["item", "delete", id]

    if archive:
        args.append("--archive")

    (status, _, stderr) = run(args, session=session, silent=True)

    if not status:
        fatal(stderr)



from typing import Optional
import keyring
import os

from ..util import KEYRING_SERVICE, is_plain_secret_storage, fatal, run, session_file


def signin(account: str) -> Optional[str]:
    (status, stdout, stderr) = run(["signin", "--raw", f"--account={account}"], silent=True)

    if not status:
        fatal(stderr)

    if is_plain_secret_storage():
        try:
            fd = os.open(session_file(account), os.O_CREAT | os.O_WRONLY, 0o600)

            with open(fd, "w") as f:
                f.write(stdout)

            return stdout
        except Exception as e:
            fatal(f"cannot write session file: {type(e).__name__}")
    else:
        try:
            keyring.set_password(KEYRING_SERVICE, account, stdout)

            return stdout
        except keyring.errors.KeyringError as e:
            fatal(f"cannot write session file: {type(e).__name__}")



from termcolor import colored


def red(string: str) -> str:
    return colored(string, "red")


def yellow(string: str) -> str:
    return colored(string, "yellow")


def blue(string: str) -> str:
    return colored(string, "blue")


def green(string: str) -> str:
    return colored(string, "green")



import inquirer  # type: ignore
import json
import sys

from ..util import exit, fatal, run
from ..colors import *


def show(session: str, to_json: bool, id: str, tags: str, fields: str, otp: bool, select: bool) -> None:
    if len(id) > 0:
        id = " ".join(id)
    else:
        (status, stdout, stderr) = run(["item", "list", "--format=json", f"--tags={tags}"], session=session, silent=True)

        if not status:
            fatal(stderr)

        entries = json.loads(stdout)

        if len(entries) == 0:
            exit("No entry matched these filters.")
        if not select and len(entries) > 1:
            print(f"{yellow('WARN')}: Multiple entries responded to this query, showing the first", file=sys.stderr)

        if not select:
            id = entries[0]["id"]
        else:
            entries = list(map(lambda e: (f"{e['vault']['name']}: {e['title']} ({e['id']})", e["id"]), sorted(entries, key=lambda entry: entry["title"])))  # type: ignore
            entry = inquirer.prompt([inquirer.List("entry", message="Select an entry", choices=entries)])

            id = entry["entry"]

    args = ["item", "get", id]

    if fields is not None:
        args.append(f"--fields={fields}")

    if otp:
        args.append("--otp")

    run(args, json=to_json, session=session)



from __future__ import annotations

import getpass
import phonenumbers
import secrets
import string
import validators  # type: ignore

from enum import Enum
from urllib.parse import urlparse

from .util import fatal


class FieldKind(Enum):
    PLAIN = "text"
    SECRET = "password"
    URL = "url"
    EMAIL = "email"
    PHONE = "phone"
    TOTP = "otp"

    def is_secret(self) -> bool:
        return self in [FieldKind.SECRET, FieldKind.TOTP]


def is_url(value: str) -> bool:
    try:
        result = urlparse(value)

        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def is_phone_number(value: str) -> bool:
    try:
        result = phonenumbers.parse(value)

        return phonenumbers.is_possible_number(result)
    except phonenumbers.NumberParseException:
        return False


class Field:
    kind: FieldKind
    key: str
    value: str

    def __init__(self, kind: FieldKind, key: str, value: str):
        self.kind = kind
        self.key = key
        self.value = value

    @classmethod
    def from_spec(cls, spec: str, length: int, symbols: bool) -> Field:
        if "=" not in spec:
            fatal(f"field spec '{spec}' is invalid")

        kind = FieldKind.PLAIN
        [key, value] = spec.split("=", maxsplit=1)

        if key.startswith("@"):
            kind = FieldKind.SECRET
            key = key.lstrip("@")
        elif key.startswith("+"):
            kind = FieldKind.TOTP
            key = key.lstrip("+")
        elif is_url(value):
            kind = FieldKind.URL
        elif validators.email(value):
            kind = FieldKind.EMAIL
        elif is_phone_number(value):
            kind = FieldKind.PHONE

        if value == "":
            if not kind.is_secret():
                kind = FieldKind.SECRET

            value = getpass.getpass(f'Secret value for "{key}": ')
        elif value == "-":
            if kind == FieldKind.TOTP:
                fatal("TOTP field value cannot be generated")

            alphabet = string.ascii_letters + string.digits

            if symbols:
                alphabet = alphabet + string.punctuation

            kind = FieldKind.SECRET
            value = "".join(secrets.choice(alphabet) for i in range(length))

        return Field(kind, key, value)

    def to_string(self) -> str:
        if self.kind == FieldKind.TOTP:
            self.value = f"otpauth://totp/?secret={self.value}"

        return f"{self.key}[{self.kind.value}]={self.value}"



import json
import sys

from ..util import exit, fatal, run, create_table


def vaults(session: str, to_json: bool) -> None:
    (status, stdout, stderr) = run(["vault", "list", f"--session={session}"], json=True, silent=True)

    if not status:
        fatal(stderr)

    if to_json:
        print(stdout)
        sys.exit(0)

    vaults = json.loads(stdout)

    if len(vaults) == 0:
        exit("There are no vaults in this account.")

    table = create_table(["ID", "NAME"])
    table.add_rows(list(map(lambda vault: [vault["id"], vault["name"]], vaults)))

    print(table)


def vault(session: str, to_json: bool, id: str) -> None:
    run(["vault", "get", id, f"--session={session}"], json=to_json)



from ..util import run


def share(session: str, id: str, time: str, once: bool) -> None:
    args = ["item", "share", id]

    if time is not None:
        args.append(f"--expiry={time}")

    if once:
        args.append("--view-once")

    run(args, session=session)



import keyring
import os
import subprocess
import sys

from os import path
from prettytable import PrettyTable  # type: ignore
from typing import Optional, List, Tuple

from .colors import *

CACHE_DIR = path.expanduser("~/.cache/1p")
KEYRING_SERVICE = "com.github.apognu.1p"


def exit(message: str) -> None:
    print(message.strip(), file=sys.stderr)
    sys.exit(0)


def fatal(message: str) -> None:
    print(f"{red('ERROR:')} {message.strip()}", file=sys.stderr)
    sys.exit(1)


def run(args: List[str], session: Optional[str] = None, json: bool = False, silent: bool = False, stdin: Optional[str] = None) -> Tuple[bool, str, str]:
    if session is not None:
        args.append(f"--session={session}")

    if json:
        args.append("--format=json")

    if stdin is not None:
        input = stdin.encode()
        pipe = subprocess.PIPE
    else:
        input = None
        pipe = None

    cmd = subprocess.Popen(["op"] + args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=pipe)
    stdout, stderr = cmd.communicate(input=input)

    if not silent:
        if len(stdout) > 0:
            print(stdout.decode("utf-8").strip())

        if len(stderr) > 0:
            print(stderr.decode("utf-8").strip())

    return (cmd.returncode == 0, stdout.decode("utf-8"), stderr.decode("utf-8"))


def is_plain_secret_storage() -> bool:
    return os.getenv("ONEP_SECRET_BACKEND") == "plain"


def init_secret_storage() -> None:
    if is_plain_secret_storage():
        if not path.isdir(CACHE_DIR):
            os.mkdir(CACHE_DIR, mode=0o700)

        if oct(os.stat(CACHE_DIR).st_mode & 0o777) != oct(0o700):
            fatal("cannot use session directory because mode is not 700")


def session_file(account: str) -> str:
    return path.join(CACHE_DIR, f"{account}.token")


def load_session(account: str) -> Optional[str]:
    if is_plain_secret_storage():
        try:
            file = session_file(account)

            if oct(os.stat(file).st_mode & 0o777) != oct(0o600):
                fatal("cannot use session because mode is not 600")

            with open(file) as f:
                return f.read()
        except FileNotFoundError:
            return None
        except Exception as e:
            fatal(f"cannot read session file: {type(e).__name__}")
    else:
        try:
            return keyring.get_password(KEYRING_SERVICE, account)
        except keyring.errors.KeyringError as e:
            fatal(f"cannot read session file: {type(e).__name__}")


def check_session(account: str) -> Optional[str]:
    from onep.commands.signin import signin

    session = load_session(account)

    if session is None:
        return signin(account)

    (status, _, _) = run(["account", "get"], session=session, silent=True)

    if not status:
        return signin(account)

    return session


def create_table(headers: List[str]) -> PrettyTable:
    table = PrettyTable()
    table.field_names = headers
    table.align = "l"
    table.padding_width = 2
    table.border = False

    return table



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['onep', 'onep.commands']

package_data = \
{'': ['*']}

install_requires = \
['inquirer>=2.9.2,<3.0.0',
 'keyring>=23.5.0,<24.0.0',
 'phonenumbers>=8.12.48,<9.0.0',
 'prettytable>=3.3.0,<4.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'validators>=0.19.0,<0.20.0']

entry_points = \
{'console_scripts': ['1p = onep.onep:main']}

setup_kwargs = {
    'name': 'onep',
    'version': '0.6.1',
    'description': '1Password CLI helper',
    'long_description': "# 1p - 1Password CLI helper\n\n`1p` is a wrapper around 1Password's `op` CLI tool to give a more user-friendly interface to secret retrieval.\n\nRequires Python >= 3.7.\n\nIt opens a session with 1Password and stores the session token in the system's keyring, optionally, you can run with `ONEP_SECRET_BACKEND=plain` to store it under `~/.cache/1p`.\n\nIt requires that the configured 1Password account has a shorthand defined (when using `op account add`).\n\n## Installation\n\n```shell\n$ # From PyPI\n$ pip install onep\n\n$ # Development build from GitHub\n$ https://github.com/apognu/1p/releases/download/tip/onep-tip-py3-none-any.whl\n```\n\n## Usage\n\n```shell\n$ 1p --help\nusage: 1p [-h] [-j] ACCOUNT COMMAND ...\n\npositional arguments:\n  ACCOUNT     shorthand of the 1Password account\n  COMMAND\n    signin    authenticate into a 1Password account\n    vaults    list available vaults\n    vault     show information about a vault\n    search    search entries matching provided term\n    show      display an entry\n    create    create an entry\n    edit      edit an entry\n    delete    delete an entry\n    download  download a document\n    upload    upload a document\n    share     get a shareable link to an item\n\noptions:\n  -h, --help  show this help message and exit\n  -j, --json  format output as JSON\n\n$ 1p personal search -t social\nID                            TITLE\n__________________________    GitHub\n__________________________    Twitter\n```\n\n## Item creation syntax\n\nItem creation syntax tries to determine the type of the provided values (URLs, email addresses and phone numbers), if possible. It also provides some utility to control the way values are entered and interpreted:\n\n * `field=` will set the type as `password` and prompt for the value interactively\n * `field=-` will set the type as `password` and generate a random value\n * `@field=value` will explicitely set the type as `password`\n * `+field=totpsecret` will consider the provided value as a TOTP secret key\n * `section.field=value` will create a field under a named section\n",
    'author': 'Antoine POPINEAU',
    'author_email': 'antoine@popineau.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/apognu/1p',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import json
import sys

from typing import List, Optional

from ..util import fatal, run
from ..fields import Field
from ..colors import *


def create(
    session: str,
    vault: str,
    category: str,
    title: str,
    specs: List[str],
    tags: Optional[str],
    length: int,
    symbols: bool,
) -> None:
    fields = list(map(lambda spec: Field.from_spec(spec, length, symbols).to_string(), specs))
    args = ["item", "create", f"--category={category}", f"--vault={vault}", f"--title={title}"] + fields

    if tags is not None:
        args.append(f"--tags={tags}")

    (status, _, stderr) = run(args, session=session, silent=True)

    if not status:
        fatal(stderr)


def edit(
    session: str,
    id: str,
    specs: List[str],
    tags: Optional[str],
    delete: Optional[List[str]],
    length: int,
    symbols: bool,
) -> None:
    fields = list(map(lambda spec: Field.from_spec(spec, length, symbols).to_string(), specs))

    if delete is not None:
        for field in delete:
            fields.append(f"{field}[delete]")

    args = ["item", "edit", id] + fields

    if tags is not None:
        args.append(f"--tags={tags}")

    (status, _, stderr) = run(args, session=session, silent=True)

    if not status:
        fatal(stderr)
